#ifndef GOAP_PLANNER_HPP
#define GOAP_PLANNER_HPP

#include "action.hpp"
#include "godot_cpp/variant/variant.hpp"
#include <godot_cpp/classes/node.hpp>
#include <godot_cpp/classes/resource.hpp>
#include <godot_cpp/templates/vector.hpp>
#include <godot_cpp/variant/string_name.hpp>

namespace godot {
class CharacterActor;

namespace goap {
class GlobalWorldState;

struct State {
    static State new_move_to(Vector3 location);
    static State new_animate(StringName animation);
    static State new_activate(Node *node);
    enum Type {
        STATE_MOVE_TO,
        STATE_ANIMATE,
        STATE_ACTIVATE
    };
    State::Type type;
    union {
        Vector3 move_to;
        StringName animate;
        Node *activate;
    };
};

class Goal : public Resource {
    GDCLASS(Goal, Resource);
    static void _bind_methods();

    void set_goal_state(Dictionary dict);
    Dictionary get_goal_state() const;
public:
    WorldState goal_state{};
};

struct PlannerNode {
    WorldState open_requirements{};
    WorldState state{};
    Ref<Action> last_edge{};

    PlannerNode() = default;
    PlannerNode(PlannerNode const &src) = default;

    static PlannerNode goal_node(WorldState const &goal);

    PlannerNode new_node_along(Ref<Action> action) const;
};

class Planner : public Node {
    GDCLASS(Planner, Node);
    static void _bind_methods();
public:
    virtual void _ready() override;
    virtual void _process(double delta_time) override;

    Array gdscript_make_plan(Ref<Goal> goal);
    Vector<Ref<Action>> make_plan(Ref<Goal> goal);

    Variant get_world_property(StringName prop_key);

    bool can_do(Ref<Action> action);
    Vector<PlannerNode> find_neighbours_of(PlannerNode &node);
    Vector<Ref<Action>> find_actions_satisfying(WorldState requirements);

    void set_actions(Array actions);
    Array get_actions() const;
private:
    CharacterActor *actor{nullptr};
    WorldState world_state_override{};
    WorldState cached_world_state{};
    GlobalWorldState *global_world_state{nullptr};

    Vector<Ref<Action>> actions{};
};

struct PlannerNodeHasher {
    static _FORCE_INLINE_ uint32_t hash(godot::goap::PlannerNode const &node) {
        uint32_t hash{1};
        for(KeyValue<StringName, Variant> const &kvp : node.state) {
            hash = hash_murmur3_one_32(kvp.key.hash(), hash);
            hash = hash_murmur3_one_32(kvp.value.hash(), hash);
        }
        return hash_fmix32(hash);
    }
};

static _FORCE_INLINE_ bool operator==(PlannerNode const &lhs, PlannerNode const &rhs) {
    return PlannerNodeHasher::hash(lhs) == PlannerNodeHasher::hash(rhs);
}
static _FORCE_INLINE_ bool operator!=(PlannerNode const &lhs, PlannerNode const &rhs) {
    return !(lhs == rhs);
}
static _FORCE_INLINE_ bool operator<(PlannerNode const &lhs, PlannerNode const &rhs) {
    return lhs.open_requirements.size() < rhs.open_requirements.size();
}
static _FORCE_INLINE_ bool operator>=(PlannerNode const &lhs, PlannerNode const &rhs) {
    return !(lhs < rhs);
}
static _FORCE_INLINE_ bool operator>(PlannerNode const &lhs, PlannerNode const &rhs) {
    return lhs.open_requirements.size() > rhs.open_requirements.size();
}
static _FORCE_INLINE_ bool operator<=(PlannerNode const &lhs, PlannerNode const &rhs) {
    return !(lhs > rhs);
}
}
}


#endif // !GOAP_PLANNER_HPP
